import { GuildPluginData } from "knub";
import { ReactionRolesPluginType } from "../types";
import { ReactionRole } from "../../../data/entities/ReactionRole";
import { TextChannel } from "eris";
import { isDiscordRESTError, sleep, isSnowflake } from "../../../utils";
import { logger } from "../../../logger";
import { LogsPlugin } from "../../Logs/LogsPlugin";
import { LogType } from "../../../data/LogType";

const CLEAR_ROLES_EMOJI = "❌";

/**
 * @return Errors encountered while applying reaction roles, if any
 */
export async function applyReactionRoleReactionsToMessage(
  pluginData: GuildPluginData<ReactionRolesPluginType>,
  channelId: string,
  messageId: string,
  reactionRoles: ReactionRole[],
): Promise<string[]> {
  const channel = pluginData.guild.channels.get(channelId) as TextChannel;
  if (!channel) return;

  const errors = [];
  const logs = pluginData.getPlugin(LogsPlugin);

  let targetMessage;
  try {
    targetMessage = await channel.getMessage(messageId);
  } catch (e) {
    if (isDiscordRESTError(e)) {
      if (e.code === 10008) {
        // Unknown message, remove reaction roles from the message
        logs.log(LogType.BOT_ALERT, {
          body: `Removed reaction roles from unknown message ${channelId}/${messageId} (${pluginData.guild.id})`,
        });
        await pluginData.state.reactionRoles.removeFromMessage(messageId);
      } else {
        logs.log(LogType.BOT_ALERT, {
          body: `Error ${e.code} when applying reaction roles to message ${channelId}/${messageId}: ${e.message}`,
        });
      }

      errors.push(`Error ${e.code} while fetching reaction role message: ${e.message}`);
      return errors;
    } else {
      throw e;
    }
  }

  // Remove old reactions, if any
  try {
    await targetMessage.removeReactions();
  } catch (e) {
    if (isDiscordRESTError(e)) {
      errors.push(`Error ${e.code} while removing old reactions: ${e.message}`);
      logs.log(LogType.BOT_ALERT, {
        body: `Error ${e.code} while removing old reaction role reactions from message ${channelId}/${messageId}: ${e.message}`,
      });
      return errors;
    }

    throw e;
  }

  await sleep(1500);

  // Add reaction role reactions
  const emojisToAdd = reactionRoles.map(rr => rr.emoji);
  emojisToAdd.push(CLEAR_ROLES_EMOJI);

  for (const rawEmoji of emojisToAdd) {
    const emoji = isSnowflake(rawEmoji) ? `foo:${rawEmoji}` : rawEmoji;

    try {
      await targetMessage.addReaction(emoji);
      await sleep(1250); // Make sure we don't hit rate limits
    } catch (e) {
      if (isDiscordRESTError(e)) {
        if (e.code === 10014) {
          pluginData.state.reactionRoles.removeFromMessage(messageId, rawEmoji);
          errors.push(`Unknown emoji: ${emoji}`);
          logs.log(LogType.BOT_ALERT, {
            body: `Could not add unknown reaction role emoji ${emoji} to message ${channelId}/${messageId}`,
          });
          continue;
        } else if (e.code === 50013) {
          errors.push(`Missing permissions to apply reactions`);
          logs.log(LogType.BOT_ALERT, {
            body: `Error ${e.code} while applying reaction role reactions to ${channelId}/${messageId}: ${e.message}`,
          });
          break;
        }
      }

      throw e;
    }
  }

  return errors;
}
